#!/usr/bin/env python3
"""
Function to access the job scheduler.
"""


def _dependency_fun(line):
    if not line:
        return []
    dependencies = line.split(",")
    dependencies = [dependency.split(":") for dependency in dependencies]
    return [{"modified": modifier, "job_id": int(job_id)} for modifier, job_id in dependencies]


_job_field_mapping = {
    "job_id": ("jobid", int),
    "node_list": ("nodelist", lambda line: line.split(",")),
    "username": ("username", lambda x: x),
    "work_dir": ("workdir", lambda x: x),
    "state": ("state", lambda x: x),
    "partition": ("partition", lambda x: x),
    "account": ("account", lambda x: x),
    "reason": ("reason", lambda x: x),
    "reservation": ("reservation", lambda x: x),
    "time_left": ("timeleft", lambda x: x),
    "time_limit": ("timelimit", lambda x: x),
    "time_used": ("timeused", lambda x: x),
    "name": ("name", lambda x: x),
    "command": ("command", lambda x: x),
    "comment": ("comment", lambda x: x),
    "dependency": ("dependency", _dependency_fun),
    "end_time": ("endtime", lambda x: x),
    "exit_code": ("exit_code", lambda line: int(line)),
    "gres": ("gres", lambda line: line.split(",")),
}


def os_call(command):
    """
    Does a subprocess call with the given command.

    Args:
        command (str): Shell command to execute.

    Returns:
        String output of the command.
    """
    import subprocess

    try:
        return subprocess.run(command,
                              shell=True,
                              check=True,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              universal_newlines=True).stdout
    except subprocess.CalledProcessError:
        raise


def get_jobs(job_filter=None, user_filter=None, partition_filter=None, line_length=200):
    filter_keys = {
        "-j": job_filter,
        "-u": user_filter,
        "-p": partition_filter,
    }

    # Build the request string
    mapping_keys, mapping_values = zip(*_job_field_mapping.items())
    format_strings = [f"{key}:{line_length}" for key, _ in mapping_values]
    format_string = ",".join(format_strings)

    # Assemble filters
    filters = []
    for filter_key, filter_values in filter_keys.items():
        if not filter_values:
            continue
        filter_values = ",".join([str(item) for item in filter_values])
        filters.append(f"{filter_key} {filter_values}")

    # Execute command
    filters = " ".join(filters)
    # -h hides the header
    command = f"squeue -h {filters} -O {format_string}"
    command_result = os_call(command)

    # Parse jobs
    results = []
    lines = command_result.split("\n")
    for line in lines:
        if not line:
            continue
        result_dict = {}
        # We cannot split the string at a space, because some may output no characters at all (e.g. dependency).
        # So we chunk the line according to the selected line length. Strip leading and trailing spaces.
        assert len(line) % line_length == 0
        line_elements = [line[i * line_length: (i + 1) * line_length].strip() for i in range(len(line) // line_length)]
        for line_element, mapping_key, mapping_value in zip(line_elements, mapping_keys, mapping_values):
            _, fn = mapping_value
            try:
                value = fn(line_element)
            except ValueError:
                value = line_element
            result_dict[mapping_key] = value
        results.append(result_dict)
    return results


def get_job_status(job_id):
    import shlex

    command = f"scontrol show job {job_id}"
    command_result = os_call(command)

    # Split using shlex, it splits at newline and space
    key_values = shlex.split(command_result)
    result_dict = {}
    for key_value in key_values:
        # Split at first "=" sign
        key, *values = key_value.split("=")
        # Reassemble value
        value = "=".join(values)
        result_dict[key] = value
    return result_dict


def submit_job(account, time, work_dir, script_path, name=None, gres=None, mail_type=None, mail_user=None, num_nodes=1, partition=None, reservation=None, dependencies=None, tasks_per_node=None):
    import os

    # Assemble command
    output_name_prefix = f"{name}_" if name else ""
    output_file = os.path.join(work_dir, f"{output_name_prefix}output_%j.out")
    error_file = os.path.join(work_dir, f"{output_name_prefix}error_%j.out")
    command = f"sbatch --account={account} --time={time} --output={output_file} --error={error_file}"

    if name:
        command += f" --job-name={name}"
    if gres:
        command += f" --gres={gres}"
    if mail_type:
        command += f" --mail-type={mail_type}"
    if mail_user:
        command += f" --mail-user={mail_user}"
    if num_nodes:
        command += f" --nodes={num_nodes}"
    if work_dir:
        command += f" --chdir={work_dir}"
    if partition:
        command += f" --partition={partition}"
    if reservation:
        command += f" --reservation={reservation}"
    if dependencies:
        dependency_string = ",".join([f"{dep['modifier']}:{dep['job_id']}" for dep in dependencies])
        command += f" --dependency={dependency_string}"
    if tasks_per_node:
        command += f" --ntasks-per-node={tasks_per_node}"

    command = f"{command} {script_path}"

    #  Execute command
    command_result = os_call(command)

    # Extract job id from result
    job_id = int(command_result.split(" ")[3])

    return job_id, command


def cancel_job(job_id):
    command = f"scancel {job_id}"
    os_call(command)
