#!/usr/bin/env python3
"""
Access to Mondo database
"""
from pymongo import MongoClient, DESCENDING
from . import models
from .config import DATABASE_HOST, DATABASE_PORT, DATABASE_NAME, DATABASE_USERNAME, DATABASE_PASSWORD

_project_schema = models.ProjectSchema()
_task_schema = models.TaskSchema()
_job_schema = models.JobSchema()


def _get_timestamp():
    from time import time
    return int(time())


def _project_to_mongodb_item(project):
    document = _project_schema.dump(project)
    return document


class ProjectNotFoundError(Exception):
    def __init__(self, project_id):
        message = f"Project not found: {project_id}"
        self.project_id = project_id
        super(ProjectNotFoundError, self).__init__(message)


class TaskNotFoundError(Exception):
    def __init__(self, project_id, task_id):
        message = f"Task not found: {project_id} - {task_id}"
        self.project_id = project_id
        self.task_id = task_id
        super(TaskNotFoundError, self).__init__(message)


class JobNotFoundError(Exception):
    def __init__(self, project_id, task_id, job_id):
        message = f"Job not found: {project_id} - {task_id} - {job_id}"
        self.project_id = project_id
        self.task_id = task_id
        self.job_id = job_id
        super(JobNotFoundError, self).__init__(message)


class ForbiddenOperationError(Exception):
    def __init__(self, message):
        self.message = message
        super(ForbiddenOperationError, self).__init__(message)


class Database(object):
    """
    Connection object to encapsulate the MongoDB database
    """

    def __init__(self):
        self.client = None

    def open(self):
        if self.client:
            raise RuntimeError("Cannot open connection: Connection already open.")
        self.client = MongoClient(DATABASE_HOST,
                                  username=DATABASE_USERNAME,
                                  password=DATABASE_PASSWORD,
                                  port=DATABASE_PORT)
        return self

    def close(self):
        if self.client is None:
            raise RuntimeError("Cannot close connection: Connection already closed.")
        self.client.close()
        self.client = None

    def __enter__(self):
        return self.open()

    def __exit__(self, *args):
        self.close()

    @property
    def database(self):
        return self.client.get_database(DATABASE_NAME)

    @property
    def projects(self):
        return self.database.projects

    @property
    def project_ids(self):
        return self.database.project_ids

    def _update_project_fields(self, project_id, key_value_dict, operation="set"):
        """
        Update a single field

        Args:
            project_id (int): Id of the project.
            key_value_dict (dict): Dictionary of keys and values to update.
        """
        forbidden_fields = ("modified", "project_id", "created")
        operations = {
            "set": "$set",
            "add": "$push",
            "delete": "$pull",
        }

        if any(field in key_value_dict for field in forbidden_fields):
            message = f"Following fields may not be updated: {', '.join(field for field in key_value_dict if field in forbidden_fields)}"
            raise ForbiddenOperationError(message)

        mongo_operation = operations[operation]
        update_res = self.projects.update_one({"project_id": project_id},
                                              {
                                                  mongo_operation: key_value_dict
                                              })
        # Raise error if no projects were affected
        # This means the project id does not exist
        if operation == "delete":
            affected_count = update_res.deleted_count
        else:
            affected_count = update_res.matched_count
        if affected_count == 0:
            raise ProjectNotFoundError(project_id=project_id)

        # Update modified date
        if operation != "delete":
            self._update_project_modified_date(project_id=project_id)

    def _update_project_modified_date(self, project_id):
        """
        Update the modified date of a project.
        """
        # Update modified data of project
        modified_date = _get_timestamp()
        self.projects.update_one({"project_id": project_id},
                                 {
                                     "$set":
                                         {
                                             "modified": str(modified_date),
                                         },
                                 })
        return modified_date

    def insert_project(self, project):
        # Find current maximum project id
        max_existing = self.project_ids.find_one(sort=[("project_id", DESCENDING)])
        if max_existing:
            new_id = max_existing["project_id"] + 1
        else:
            new_id = 1
        # Insert new id into project
        project.project_id = new_id
        # Insert created and modified date
        project.created = project.modified = _get_timestamp()
        # If the project has tasks, give them ids and modified/create dates
        if project.tasks:
            for i, task in enumerate(project.tasks, 1):
                task.created = task.modified = project.created
                task.task_id = i
        # Serialize the project, while taking care that the id is correctly used by MongoDB
        document = _project_to_mongodb_item(project=project)
        # Insert into database
        self.projects.insert_one(document)
        # Insert project id
        self.project_ids.insert_one({"project_id": new_id, "next_task_id": 1, })

        return document

    def update_project_values(self, project_id, key_value_dict):
        self._update_project_fields(project_id=project_id,
                                    key_value_dict=key_value_dict,
                                    operation="set")

    def get_project_by_id(self, project_id):
        """
        Get a project with a given id.

        Args:
            project_id (int): Id of the project.

        Returns:
            atlas_server.models.Project
        """
        project = self.projects.find_one({"project_id": project_id})
        if not project:
            raise ProjectNotFoundError(project_id=project_id)
        return _project_schema.load(project, unknown=True)

    def get_projects_for_user(self, username):
        projects = list(self.projects.find({"owner": username}))
        # noinspection PyTypeChecker
        return _project_schema.load(projects, many=True, unknown=True)

    def delete_project(self, project_id):
        deleted_res = self.projects.delete_one({"project_id": project_id})
        if deleted_res.deleted_count == 0:
            raise ProjectNotFoundError(project_id=project_id)

    def _update_task_fields(self, project_id, task_id, key_value_dict, operation="set"):
        """
        Update a single field

        Args:
            project_id (int): Id of the project.
            key_value_dict (dict): Dictionary of keys and values to update.
        """
        forbidden_fields = ("modified", "task_id", "created")
        operations = {
            "set": "$set",
            "add": "$push",
        }

        if any(field in key_value_dict for field in forbidden_fields):
            message = f"Following fields may not be updated: {', '.join(field for field in key_value_dict if field in forbidden_fields)}"
            raise ForbiddenOperationError(message)

        mongo_operation = operations[operation]
        # Modify the update dict to match the update syntax. We prefix "tasks.$." before each key
        update_dict = {f"tasks.$.{key}": value for key, value in key_value_dict.items()}
        update_res = self.projects.update_one({"project_id": project_id, "tasks.task_id": task_id, }, {mongo_operation: update_dict})
        # Raise error if no projects were affected
        # This means the project id does not exist
        affected_count = update_res.modified_count
        if affected_count == 0:
            raise TaskNotFoundError(project_id=project_id, task_id=task_id)

        # Update modified date (if we did not delete the task)
        self._update_task_modified_date(project_id=project_id, task_id=task_id)

    def _delete_task(self, project_id, task_id):
        selection_dict = {"project_id": project_id, "tasks.task_id": task_id}
        delete_dict = {"$pull": {"tasks": {"task_id": task_id}}}
        update_res = self.projects.update_one(selection_dict, delete_dict)
        affected_count = update_res.modified_count
        if affected_count == 0:
            raise TaskNotFoundError(project_id=project_id, task_id=task_id)

    def _update_task_modified_date(self, project_id, task_id):
        """
        Update the modified date of a task.
        """
        # Update modified data of project
        modified_date = _get_timestamp()
        self.projects.update_one({"project_id": project_id, "tasks.task_id": task_id, },
                                 {
                                     "$set":
                                         {
                                             "task.$.modified": str(modified_date),
                                         },
                                 })
        return modified_date

    # noinspection PyUnresolvedReferences
    def add_task_to_project(self, project_id, task):
        # Get the project and find out how many tasks it already has
        project = self.get_project_by_id(project_id=project_id)
        if project is None:
            raise ProjectNotFoundError(project_id)
        project_ids = self.project_ids.find_one({"project_id": project_id})
        if not project_ids:
            # Fallback for projects which have no entry in project_ids
            if not project.tasks:
                new_task_id = 1
            else:
                new_task_id = max(task.task_id for task in project.tasks) + 1
            project_ids.insert_one({"project_id": project_id, "next_task_id": new_task_id})
        else:
            new_task_id = project_ids["next_task_id"]

        task.task_id = new_task_id
        # Add created and modified info to task
        task.created = _get_timestamp()
        task.modified = _get_timestamp()
        # Serialize task
        document = _task_schema.dump(task)
        # Insert task into the projects task list
        self._update_project_fields(project_id=project_id,
                                    key_value_dict={
                                        "tasks": document
                                    },
                                    operation="add")
        # Update next task id
        self.project_ids.update_one({"project_id": project_id}, {"$set": {"next_task_id": new_task_id + 1}})
        return document

    def get_task_by_id(self, project_id, task_id):
        res = self.projects.find_one(
            {
                "project_id": project_id,
                "tasks.task_id": task_id
            },
            # Projection, only get the specific task
            {
                "tasks.$": 1,
            }
        )
        if not res:
            raise TaskNotFoundError(project_id=project_id, task_id=task_id)
        return _task_schema.load(res["tasks"][0])

    def update_task_values(self, project_id, task_id, key_value_dict, operation="set"):
        self._update_task_fields(project_id=project_id,
                                 task_id=task_id,
                                 key_value_dict=key_value_dict,
                                 operation=operation)

    def delete_task(self, project_id, task_id):
        # delete_res = self.projects.
        self._delete_task(project_id=project_id,
                          task_id=task_id)

    def get_job_by_id(self, project_id, task_id, job_id):
        # We cannot do double nested updates with mongoDB, so we get the task
        # and search for the task in the result set
        try:
            task = self.get_task_by_id(project_id=project_id,
                                       task_id=task_id)
        except TaskNotFoundError:
            raise JobNotFoundError(project_id=project_id,
                                   task_id=task_id,
                                   job_id=job_id)
            # Search for the job
        # noinspection PyUnresolvedReferences
        job = [job for job in task.jobs if job.job_id == job_id]
        if not job:
            raise JobNotFoundError(project_id=project_id,
                                   task_id=task_id,
                                   job_id=job_id)
